@@ -14,6 +14,7 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <sys/queue.h>
 #include <sys/types.h>
 
 #include <ctype.h>
@@ -444,9 +445,8 @@ done:
 __dead static void
 usage_pack(void)
 {
-	fprintf(stderr, "usage: %s pack [-a] [-r repository-path] "
-	    "[-x reference] [-q] [reference ...]\n",
-	    getprogname());
+	fprintf(stderr, "usage: %s pack [-aq] [-r repository-path] "
+	    "[-x reference] [reference ...]\n", getprogname());
 	exit(1);
 }
 
@@ -924,7 +924,7 @@ done:
 __dead static void
 usage_listpack(void)
 {
-	fprintf(stderr, "usage: %s listpack [-h] [-s] packfile-path\n",
+	fprintf(stderr, "usage: %s listpack [-hs] packfile-path\n",
 	    getprogname());
 	exit(1);
 }
@@ -980,7 +980,7 @@ list_pack_cb(void *arg, struct got_object_id *id, int type, off_t offset,
 		break;
 	case GOT_OBJ_TYPE_REF_DELTA:
 		type_str = "ref-delta";
-		err = got_object_id_str(&base_id_str, base_id);	
+		err = got_object_id_str(&base_id_str, base_id);
 		if (err)
 			goto done;
 		if (asprintf(&delta_str, " base-id %s", base_id_str) == -1) {
@@ -1115,8 +1115,8 @@ done:
 __dead static void
 usage_cleanup(void)
 {
-	fprintf(stderr, "usage: %s cleanup [-a] [-p] [-n] [-r repository-path] "
-	    "[-q]\n", getprogname());
+	fprintf(stderr, "usage: %s cleanup [-anpq] [-r repository-path]\n",
+	    getprogname());
 	exit(1);
 }
 
