@@ -16,6 +16,7 @@
  */
 
 
+#include <sys/queue.h>
 #include <ctype.h>
 #include <errno.h>
 #include <stdint.h>
@@ -603,7 +604,8 @@ diff_main(const struct diff_config *config, struct diff_data *left,
 
 	struct diff_state state = {
 		.result = result,
-		.recursion_depth_left = config->max_recursion_depth ? : UINT_MAX,
+		.recursion_depth_left = config->max_recursion_depth ?
+		    config->max_recursion_depth : UINT_MAX,
 		.kd_buf = NULL,
 		.kd_buf_size = 0,
 	};
