@@ -16,6 +16,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/queue.h>
 #include <sys/uio.h>
 #include <sys/socket.h>
 #include <sys/wait.h>
@@ -175,7 +176,7 @@ pack_child_send_tempfiles(struct imsgbuf *ibuf, struct got_pack *pack)
 	const struct got_error *err;
 	int basefd = -1, accumfd = -1;
 
-	/* 
+	/*
 	 * For performance reasons, the child will keep reusing the
 	 * same temporary files during every object request.
 	 * Opening and closing new files for every object request is
@@ -886,7 +887,7 @@ got_object_id_queue_copy(const struct got_object_id_queue *src,
 		 * Deep-copy the object ID only. Let the caller deal
 		 * with setting up the new->data pointer if needed.
 		 */
-		err = got_object_qid_alloc(&new, &qid->id); 
+		err = got_object_qid_alloc(&new, &qid->id);
 		if (err) {
 			got_object_id_queue_free(dest);
 			return err;
@@ -2190,6 +2191,7 @@ resolve_symlink(char **link_target, const char *path,
 		}
 	}
 done:
+	free(parent_path);
 	free(tree_obj_id);
 	if (tree)
 		got_object_tree_close(tree);
@@ -2350,3 +2352,12 @@ done:
 	free(path_packfile);
 	return err;
 }
+
+const struct got_error *
+got_object_commit_dup(struct got_commit_object **ret,
+    struct got_commit_object *commit)
+{
+	*ret = commit;
+	commit->refcnt++;
+	return NULL;
+}
