@@ -16,6 +16,7 @@
  */
 
 #include <sys/types.h>
+#include <sys/queue.h>
 #include <sys/uio.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -23,6 +24,7 @@
 
 #include <errno.h>
 #include <stdint.h>
+#include <inttypes.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -495,7 +497,7 @@ add_meta(struct got_pack_meta *m, struct got_pack_metavec *v)
 		if (new == NULL)
 			return got_error_from_errno("reallocarray");
 		v->meta = new;
-		v->metasz = newsize; 
+		v->metasz = newsize;
 	}
 
 	v->meta[v->nmeta++] = m;
@@ -1371,7 +1373,7 @@ recv_painted_commit(void *arg, struct got_object_id *id, intptr_t color)
 	default:
 		/* should not happen */
 		return got_error_fmt(GOT_ERR_NOT_IMPL,
-		    "%s invalid commit color %d", __func__, color);
+		    "%s invalid commit color %"PRIdPTR, __func__, color);
 	}
 
 	STAILQ_FOREACH_SAFE(qid, a->ids, entry, tmp) {
@@ -1615,7 +1617,8 @@ paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 		default:
 			/* should not happen */
 			err = got_error_fmt(GOT_ERR_NOT_IMPL,
-			    "%s invalid commit color %d", __func__, color);
+			    "%s invalid commit color %"PRIdPTR, __func__,
+			    color);
 			goto done;
 		}
 
@@ -1737,7 +1740,7 @@ findtwixt(struct got_object_id ***res, int *nres, int *ncolored,
 		err = queue_commit_or_tag_id(id, COLOR_KEEP, &ids, repo);
 		if (err)
 			goto done;
-	}		
+	}
 
 	for (i = 0; i < ntail; i++) {
 		struct got_object_id *id = tail[i];
