@@ -21,6 +21,7 @@
  */
 
 #include <sys/types.h>
+#include <sys/queue.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/uio.h>
@@ -311,6 +312,35 @@ done:
 	return err;
 }
 
+static void
+reverse_patch(struct got_patch *p)
+{
+	struct got_patch_hunk *h;
+	size_t i;
+	int tmp;
+
+	STAILQ_FOREACH(h, &p->head, entries) {
+		tmp = h->old_from;
+		h->old_from = h->new_from;
+		h->new_from = tmp;
+
+		tmp = h->old_lines;
+		h->old_lines = h->new_lines;
+		h->new_lines = tmp;
+
+		tmp = h->old_nonl;
+		h->old_nonl = h->new_nonl;
+		h->new_nonl = tmp;
+
+		for (i = 0; i < h->len; ++i) {
+			if (*h->lines[i] == '+')
+				*h->lines[i] = '-';
+			else if (*h->lines[i] == '-')
+				*h->lines[i] = '+';
+		}
+	}
+}
+
 /*
  * Copy data from orig starting at copypos until pos into tmp.
  * If pos is -1, copy until EOF.
@@ -344,6 +374,8 @@ copy(FILE *tmp, FILE *orig, off_t copypos, off_t pos)
 	return NULL;
 }
 
+static int linecmp(const char *, const char *, int *);
+
 static const struct got_error *
 locate_hunk(FILE *orig, struct got_patch_hunk *h, off_t *pos, int *lineno)
 {
@@ -353,7 +385,7 @@ locate_hunk(FILE *orig, struct got_patch_hunk *h, off_t *pos, int *lineno)
 	size_t linesize = 0;
 	ssize_t linelen;
 	off_t match = -1;
-	int match_lineno = -1;
+	int mangled = 0, match_lineno = -1;
 
 	for (;;) {
 		linelen = getline(&line, &linesize, orig);
@@ -368,8 +400,8 @@ locate_hunk(FILE *orig, struct got_patch_hunk *h, off_t *pos, int *lineno)
 			line[linelen - 1] = '\0';
 		(*lineno)++;
 
-		if ((mode == ' ' && !strcmp(h->lines[0] + 1, line)) ||
-		    (mode == '-' && !strcmp(h->lines[0] + 1, line)) ||
+		if ((mode == ' ' && !linecmp(h->lines[0] + 1, line, &mangled)) ||
+		    (mode == '-' && !linecmp(h->lines[0] + 1, line, &mangled)) ||
 		    (mode == '+' && *lineno == h->old_from)) {
 			match = ftello(orig);
 			if (match == -1) {
@@ -380,8 +412,11 @@ locate_hunk(FILE *orig, struct got_patch_hunk *h, off_t *pos, int *lineno)
 			match_lineno = (*lineno)-1;
 		}
 
-		if (*lineno >= h->old_from && match != -1)
+		if (*lineno >= h->old_from && match != -1) {
+			if (mangled)
+				h->ws_mangled = 1;
 			break;
+		}
 	}
 
 	if (err == NULL) {
@@ -693,11 +728,56 @@ done:
 	return err;
 }
 
+static const struct got_error *
+prepare_merge(int *do_merge, char **apath, FILE **afile,
+    struct got_worktree *worktree, struct got_repository *repo,
+    struct got_patch *p, struct got_object_id *commit_id,
+    struct got_tree_object *tree, const char *path)
+{
+	const struct got_error *err = NULL;
+
+	*do_merge = 0;
+	*apath = NULL;
+	*afile = NULL;
+
+	/* don't run the diff3 merge on creations/deletions */
+	if (p->old == NULL || p->new == NULL)
+		return NULL;
+
+	if (commit_id) {
+		struct got_object_id *id;
+
+		err = got_object_tree_find_path(&id, NULL, repo, tree, path);
+		if (err)
+			return err;
+		got_sha1_digest_to_str(id->sha1, p->blob, sizeof(p->blob));
+		got_sha1_digest_to_str(commit_id->sha1, p->cid, sizeof(p->cid));
+		free(id);
+		err = open_blob(apath, afile, p->blob, repo);
+		*do_merge = err == NULL;
+	} else if (*p->blob != '\0') {
+		err = open_blob(apath, afile, p->blob, repo);
+		/*
+		 * ignore failures to open this blob, we might have
+		 * parsed gibberish.
+		 */
+		if (err && !(err->code == GOT_ERR_ERRNO && errno == ENOENT) &&
+		    err->code != GOT_ERR_NO_OBJ)
+			return err;
+		*do_merge = err == NULL;
+		err = NULL;
+	}
+
+	return err;
+}
+
 static const struct got_error *
 apply_patch(int *overlapcnt, struct got_worktree *worktree,
     struct got_repository *repo, struct got_fileindex *fileindex,
     const char *old, const char *new, struct got_patch *p, int nop,
-    struct patch_args *pa, got_cancel_cb cancel_cb, void *cancel_arg)
+    int reverse, struct got_object_id *commit_id,
+    struct got_tree_object *tree, struct patch_args *pa,
+    got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err = NULL;
 	struct stat sb;
@@ -713,20 +793,13 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 
 	*overlapcnt = 0;
 
-	/* don't run the diff3 merge on creations/deletions */
-	if (*p->blob != '\0' && p->old != NULL && p->new != NULL) {
-		err = open_blob(&apath, &afile, p->blob, repo);
-		/*
-		 * ignore failures to open this blob, we might have
-		 * parsed gibberish.
-		 */
-		if (err && !(err->code == GOT_ERR_ERRNO && errno == ENOENT) &&
-		    err->code != GOT_ERR_NO_OBJ)
-			return err;
-		else if (err == NULL)
-			do_merge = 1;
-		err = NULL;
-	}
+	err = prepare_merge(&do_merge, &apath, &afile, worktree, repo, p,
+	    commit_id, tree, old);
+	if (err)
+		return err;
+
+	if (reverse && !do_merge)
+		reverse_patch(p);
 
 	if (asprintf(&oldpath, "%s/%s", got_worktree_get_root_path(worktree),
 	    old) == -1) {
@@ -805,6 +878,19 @@ apply_patch(int *overlapcnt, struct got_worktree *worktree,
 			goto done;
 		}
 
+		if (reverse) {
+			char *s;
+			FILE *t;
+
+			s = anclabel;
+			anclabel = newlabel;
+			newlabel = s;
+
+			t = afile;
+			afile = tmpfile;
+			tmpfile = t;
+		}
+
 		err = got_opentemp_named(&mergepath, &mergefile, template);
 		if (err)
 			goto done;
@@ -906,42 +992,16 @@ done:
 	return err;
 }
 
-static void
-reverse_patch(struct got_patch *p)
-{
-	struct got_patch_hunk *h;
-	size_t i;
-	int tmp;
-
-	STAILQ_FOREACH(h, &p->head, entries) {
-		tmp = h->old_from;
-		h->old_from = h->new_from;
-		h->new_from = tmp;
-
-		tmp = h->old_lines;
-		h->old_lines = h->new_lines;
-		h->new_lines = tmp;
-
-		tmp = h->old_nonl;
-		h->old_nonl = h->new_nonl;
-		h->new_nonl = tmp;
-
-		for (i = 0; i < h->len; ++i) {
-			if (*h->lines[i] == '+')
-				*h->lines[i] = '-';
-			else if (*h->lines[i] == '-')
-				*h->lines[i] = '+';
-		}
-	}
-}
-
 const struct got_error *
 got_patch(int fd, struct got_worktree *worktree, struct got_repository *repo,
-    int nop, int strip, int reverse, got_patch_progress_cb progress_cb,
-    void *progress_arg, got_cancel_cb cancel_cb, void *cancel_arg)
+    int nop, int strip, int reverse, struct got_object_id *commit_id,
+    got_patch_progress_cb progress_cb, void *progress_arg,
+    got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err = NULL, *complete_err = NULL;
 	struct got_fileindex *fileindex = NULL;
+	struct got_commit_object *commit = NULL;
+	struct got_tree_object *tree = NULL;
 	char *fileindex_path = NULL;
 	char *oldpath, *newpath;
 	struct imsgbuf *ibuf;
@@ -987,6 +1047,16 @@ got_patch(int fd, struct got_worktree *worktree, struct got_repository *repo,
 	if (err)
 		goto done;
 
+	if (commit_id) {
+		err = got_object_open_as_commit(&commit, repo, commit_id);
+		if (err)
+			goto done;
+
+		err = got_object_open_as_tree(&tree, repo, commit->tree_id);
+		if (err)
+			goto done;
+	}
+
 	while (!done && err == NULL) {
 		struct got_patch p;
 		struct patch_args pa;
@@ -999,15 +1069,12 @@ got_patch(int fd, struct got_worktree *worktree, struct got_repository *repo,
 		if (err || done)
 			break;
 
-		if (reverse)
-			reverse_patch(&p);
-
 		err = got_worktree_patch_check_path(p.old, p.new, &oldpath,
 		    &newpath, worktree, repo, fileindex);
 		if (err == NULL)
 			err = apply_patch(&overlapcnt, worktree, repo,
-			    fileindex, oldpath, newpath, &p, nop, &pa,
-			    cancel_cb, cancel_arg);
+			    fileindex, oldpath, newpath, &p, nop, reverse,
+			    commit_id, tree, &pa, cancel_cb, cancel_arg);
 		if (err != NULL) {
 			failed = 1;
 			/* recoverable errors */
@@ -1037,6 +1104,10 @@ done:
 	if (complete_err && err == NULL)
 		err = complete_err;
 	free(fileindex_path);
+	if (tree)
+		got_object_tree_close(tree);
+	if (commit)
+		got_object_commit_close(commit);
 	if (fd != -1 && close(fd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	if (ibuf != NULL)
