@@ -15,6 +15,7 @@
  */
 
 #include <sys/types.h>
+#include <sys/queue.h>
 #include <sys/uio.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -246,22 +247,32 @@ const struct got_error *
 got_repo_pack_fds_open(int **pack_fds)
 {
 	const struct got_error *err = NULL;
-	int i, pack_fds_tmp[GOT_PACK_NUM_TEMPFILES];
+	int i;
 
 	*pack_fds = calloc(GOT_PACK_NUM_TEMPFILES, sizeof(**pack_fds));
 	if (*pack_fds == NULL)
 		return got_error_from_errno("calloc");
 
+	/*
+	 * got_repo_pack_fds_close will try to close all of the
+	 * GOT_PACK_NUM_TEMPFILES fds, even the ones that didn't manage to get
+	 * a value from got_opentempfd(), which would result in a close(0) if
+	 * we do not initialize to -1 here.
+	 */
+	for (i = 0; i < GOT_PACK_NUM_TEMPFILES; i++)
+		(*pack_fds)[i] = -1;
+
 	for (i = 0; i < GOT_PACK_NUM_TEMPFILES; i++) {
-		pack_fds_tmp[i] = got_opentempfd();
-		if (pack_fds_tmp[i] == -1) {
+		(*pack_fds)[i] = got_opentempfd();
+		if ((*pack_fds)[i] == -1) {
 			err = got_error_from_errno("got_opentempfd");
-			got_repo_pack_fds_close(pack_fds_tmp);
+			got_repo_pack_fds_close(*pack_fds);
+			*pack_fds = NULL;
 			return err;
 		}
 	}
-	memcpy(*pack_fds, pack_fds_tmp, sizeof(pack_fds_tmp));
-	return err;
+
+	return NULL;
 }
 
 const struct got_error *
@@ -658,6 +669,29 @@ read_gitconfig(struct got_repository *repo, const char *global_gitconfig_path)
 	    repo_gitconfig_path);
 	if (err)
 		goto done;
+
+	if (getenv("GOT_IGNORE_GITCONFIG") != NULL) {
+		int i;
+
+		for (i = 0; i < repo->ngitconfig_remotes; i++) {
+			got_repo_free_remote_repo_data(
+			    &repo->gitconfig_remotes[i]);
+		}
+		free(repo->gitconfig_remotes);
+		repo->gitconfig_remotes = NULL;
+		repo->ngitconfig_remotes = 0;
+
+		free(repo->gitconfig_author_name);
+		repo->gitconfig_author_name = NULL;
+		free(repo->gitconfig_author_email);
+		repo->gitconfig_author_email = NULL;
+
+		free(repo->global_gitconfig_author_name);
+		repo->global_gitconfig_author_name = NULL;
+		free(repo->global_gitconfig_author_email);
+		repo->global_gitconfig_author_email = NULL;
+	}
+
 done:
 	free(repo_gitconfig_path);
 	return err;
@@ -782,8 +816,10 @@ got_repo_open(struct got_repository **repop, const char *path,
 	err = read_gitconfig(repo, global_gitconfig_path);
 	if (err)
 		goto done;
-	if (repo->gitconfig_repository_format_version != 0)
+	if (repo->gitconfig_repository_format_version != 0) {
 		err = got_error_path(path, GOT_ERR_GIT_REPO_FORMAT);
+		goto done;
+	}
 	for (i = 0; i < repo->nextensions; i++) {
 		char *ext = repo->extensions[i];
 		int j, supported = 0;
@@ -827,6 +863,7 @@ got_repo_close(struct got_repository *repo)
 	    &repo->packidx_bloom_filters))) {
 		RB_REMOVE(got_packidx_bloom_filter_tree,
 		    &repo->packidx_bloom_filters, bf);
+		bloom_free(bf->bloom);
 		free(bf->bloom);
 		free(bf);
 	}
@@ -1122,7 +1159,7 @@ add_packidx_bloom_filter(struct got_repository *repo,
 		free(bf);
 		return got_error_from_errno("calloc");
 	}
-	
+
 	len = strlcpy(bf->path, path_packidx, sizeof(bf->path));
 	if (len >= sizeof(bf->path)) {
 		free(bf->bloom);
@@ -1236,6 +1273,7 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 	struct dirent *dent;
 	char *path_packidx = NULL;
 	int packdir_fd;
+	struct stat sb;
 
 	packdir_fd = openat(got_repo_get_fd(repo),
 	    GOT_OBJECTS_PACK_DIR, O_DIRECTORY | O_CLOEXEC);
@@ -1251,6 +1289,12 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 		goto done;
 	}
 
+	if (fstat(packdir_fd, &sb) == -1) {
+		err = got_error_from_errno("fstat");
+		goto done;
+	}
+	repo->pack_path_mtime = sb.st_mtime;
+
 	while ((dent = readdir(packdir)) != NULL) {
 		if (!got_repo_is_packidx_filename(dent->d_name,
 		    strlen(dent->d_name)))
@@ -1392,7 +1436,7 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 			return got_error_from_errno("ftruncate");
 		if (ftruncate(repo->packs[i].accumfd, 0L) == -1)
 			return got_error_from_errno("ftruncate");
-		memcpy(&tmp, &repo->packs[i], sizeof(tmp)); 
+		memcpy(&tmp, &repo->packs[i], sizeof(tmp));
 		memcpy(&repo->packs[i], &repo->packs[0],
 		    sizeof(repo->packs[i]));
 		memcpy(&repo->packs[0], &tmp, sizeof(repo->packs[0]));
@@ -1563,16 +1607,47 @@ got_repo_init(const char *repo_path)
 	return NULL;
 }
 
+static void
+purge_packidx_paths(struct got_pathlist_head *packidx_paths)
+{
+	struct got_pathlist_entry *pe;
+
+	while (!TAILQ_EMPTY(packidx_paths)) {
+		pe = TAILQ_FIRST(packidx_paths);
+		TAILQ_REMOVE(packidx_paths, pe, entry);
+		free((char *)pe->path);
+		free(pe);
+	}
+}
+
 static const struct got_error *
 match_packed_object(struct got_object_id **unique_id,
     struct got_repository *repo, const char *id_str_prefix, int obj_type)
 {
 	const struct got_error *err = NULL;
+	char *objects_pack_dir = NULL;
 	struct got_object_id_queue matched_ids;
 	struct got_pathlist_entry *pe;
+	struct stat sb;
 
 	STAILQ_INIT(&matched_ids);
 
+	objects_pack_dir = got_repo_get_path_objects_pack(repo);
+	if (objects_pack_dir == NULL)
+		return got_error_from_errno("got_repo_get_path_objects_pack");
+
+	if (stat(objects_pack_dir, &sb) == -1) {
+		if (errno != ENOENT) {
+			err = got_error_from_errno2("stat", objects_pack_dir);
+			goto done;
+		}
+	} else if (sb.st_mtime != repo->pack_path_mtime) {
+		purge_packidx_paths(&repo->packidx_paths);
+		err = got_repo_list_packidx(&repo->packidx_paths, repo);
+		if (err)
+			goto done;
+	}
+
 	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
 		const char *path_packidx = pe->path;
 		struct got_packidx *packidx;
@@ -1619,6 +1694,7 @@ match_packed_object(struct got_object_id **unique_id,
 		}
 	}
 done:
+	free(objects_pack_dir);
 	got_object_id_queue_free(&matched_ids);
 	if (err) {
 		free(*unique_id);
@@ -1633,7 +1709,7 @@ match_loose_object(struct got_object_id **unique_id, const char *path_objects,
     struct got_repository *repo)
 {
 	const struct got_error *err = NULL;
-	char *path;
+	char *path, *id_str = NULL;
 	DIR *dir = NULL;
 	struct dirent *dent;
 	struct got_object_id id;
@@ -1653,9 +1729,11 @@ match_loose_object(struct got_object_id **unique_id, const char *path_objects,
 		goto done;
 	}
 	while ((dent = readdir(dir)) != NULL) {
-		char *id_str;
 		int cmp;
 
+		free(id_str);
+		id_str = NULL;
+
 		if (strcmp(dent->d_name, ".") == 0 ||
 		    strcmp(dent->d_name, "..") == 0)
 			continue;
@@ -1673,10 +1751,8 @@ match_loose_object(struct got_object_id **unique_id, const char *path_objects,
 		 * sorted order, so we must iterate over all of them.
 		 */
 		cmp = strncmp(id_str, id_str_prefix, strlen(id_str_prefix));
-		if (cmp != 0) {
-			free(id_str);
+		if (cmp != 0)
 			continue;
-		}
 
 		if (*unique_id == NULL) {
 			if (obj_type != GOT_OBJ_TYPE_ANY) {
@@ -1691,14 +1767,12 @@ match_loose_object(struct got_object_id **unique_id, const char *path_objects,
 			*unique_id = got_object_id_dup(&id);
 			if (*unique_id == NULL) {
 				err = got_error_from_errno("got_object_id_dup");
-				free(id_str);
 				goto done;
 			}
 		} else {
 			if (got_object_id_cmp(*unique_id, &id) == 0)
 				continue; /* both packed and loose */
 			err = got_error(GOT_ERR_AMBIGUOUS_ID);
-			free(id_str);
 			goto done;
 		}
 	}
@@ -1709,6 +1783,7 @@ done:
 		free(*unique_id);
 		*unique_id = NULL;
 	}
+	free(id_str);
 	free(path);
 	return err;
 }
@@ -1718,21 +1793,25 @@ got_repo_match_object_id_prefix(struct got_object_id **id,
     const char *id_str_prefix, int obj_type, struct got_repository *repo)
 {
 	const struct got_error *err = NULL;
-	char *path_objects = got_repo_get_path_objects(repo);
-	char *object_dir = NULL;
+	char *path_objects = NULL, *object_dir = NULL;
 	size_t len;
 	int i;
 
 	*id = NULL;
 
+	path_objects = got_repo_get_path_objects(repo);
+
 	len = strlen(id_str_prefix);
-	if (len > SHA1_DIGEST_STRING_LENGTH - 1)
-		return got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
+	if (len > SHA1_DIGEST_STRING_LENGTH - 1) {
+		err = got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
+		goto done;
+	}
 
 	for (i = 0; i < len; i++) {
 		if (isxdigit((unsigned char)id_str_prefix[i]))
 			continue;
-		return got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
+		err = got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
+		goto done;
 	}
 
 	if (len >= 2) {
@@ -1768,6 +1847,7 @@ got_repo_match_object_id_prefix(struct got_object_id **id,
 		goto done;
 	}
 done:
+	free(path_objects);
 	free(object_dir);
 	if (err) {
 		free(*id);
@@ -1833,12 +1913,10 @@ got_repo_match_object_id(struct got_object_id **id, char **label,
 			return err;
 	}
 
-	err = got_repo_match_object_id_prefix(id, id_str, obj_type, repo);
-	if (err) {
-		if (err->code != GOT_ERR_BAD_OBJ_ID_STR)
-			return err;
-		err = got_ref_open(&ref, repo, id_str, 0);
-		if (err != NULL)
+	err = got_ref_open(&ref, repo, id_str, 0);
+	if (err == NULL) {
+		err = got_ref_resolve(id, repo, ref);
+		if (err)
 			goto done;
 		if (label) {
 			*label = strdup(got_ref_get_name(ref));
@@ -1847,12 +1925,23 @@ got_repo_match_object_id(struct got_object_id **id, char **label,
 				goto done;
 			}
 		}
-		err = got_ref_resolve(id, repo, ref);
-	} else if (label) {
-		err = got_object_id_str(label, *id);
-		if (*label == NULL) {
-			err = got_error_from_errno("strdup");
+	} else {
+		if (err->code != GOT_ERR_NOT_REF &&
+		    err->code != GOT_ERR_BAD_REF_NAME)
 			goto done;
+		err = got_repo_match_object_id_prefix(id, id_str,
+		    obj_type, repo);
+		if (err) {
+			if (err->code == GOT_ERR_BAD_OBJ_ID_STR)
+				err = got_error_not_ref(id_str);
+			goto done;
+		}
+		if (label) {
+			err = got_object_id_str(label, *id);
+			if (*label == NULL) {
+				err = got_error_from_errno("strdup");
+				goto done;
+			}
 		}
 	}
 done:
@@ -2014,7 +2103,7 @@ import_subdir(struct got_tree_entry **new_te, struct dirent *de,
 	if (err)
 		goto done;
 	memcpy(&(*new_te)->id, id, sizeof((*new_te)->id));
-	
+
 done:
 	free(id);
 	free(subdirpath);
